from poglink.utils import parse_list, setup_argparse


def test_parse_list():
    assert parse_list("this,       is,  a  , test    ") == ["this", "is", "a", "test"]
    assert parse_list("") == [""]
    assert parse_list("this,has,a,double,,comma") == [
        "this",
        "has",
        "a",
        "double",
        "",
        "comma",
    ]


def test_parser():
    parser = setup_argparse()

    args = parser.parse_args(
        [
            "-d",
            "~/data-dir",
            "-t",
            "MYTOKEN",
            "-p",
            "500",
            "-a",
            "test,admin",
            "--rates-urls",
            "www.test.com,www.test123.com",
            "--rates-channel-id",
            "1234",
            # "--bans-url",  # TODO: Reimplement when bans are enabled
            # "www.bans.com",# TODO: Reimplement when bans are enabled
            # "--bans-channel-id", # TODO: Reimplement when bans are enabled
            # "4321", # TODO: Reimplement when bans are enabled
            "--debug",
        ]
    )

    assert args.data_dir == "~/data-dir"
    assert args.token == "MYTOKEN"
    assert args.polling_delay == 500
    assert args.allowed_roles == ["test", "admin"]
    assert args.rates_urls == ["www.test.com", "www.test123.com"]
    assert args.rates_channel_id == "1234"
    # assert args.bans_url == "www.bans.com" # TODO: Reimplement when bans are enabled
    # assert args.bans_channel_id == "4321" # TODO: Reimplement when bans are enabled
    assert args.debug == True



import os

import pytest

from poglink.cogs.rates import Rates
from poglink.error import RatesFetchError, RatesProcessError, RatesWriteError
from poglink.models.rates import RatesDiffItem


@pytest.fixture()
def rates_cog(sample_bot):
    rates_cog = Rates(sample_bot)
    yield rates_cog

    # Clean up
    for f in rates_cog.output_paths:
        try:
            os.remove(f)
        except:
            pass


@pytest.mark.asyncio
async def test_rates_compare_posted_rates(rates_cog, rates_url_1, rates_url_2):
    output_path = rates_cog.output_paths[0]

    # First request; no changes expected
    rates_diff = await rates_cog.compare_posted_rates(rates_url_1, output_path)
    assert len(rates_diff.items) == 0

    # 2nd request; changes should be returned
    rates_diff = await rates_cog.compare_posted_rates(rates_url_2, output_path)

    assert len(rates_diff.items) == 3
    assert (
        RatesDiffItem(key="HarvestAmountMultiplier", old_val="3.0", new_val="2.0")
        in rates_diff.items
    )
    assert (
        RatesDiffItem(key="MatingIntervalMultiplier", old_val="0.6", new_val="0.7")
        in rates_diff.items
    )
    assert (
        RatesDiffItem(key="HexagonRewardMultiplier", old_val="1.5", new_val="2.5")
        in rates_diff.items
    )

    # 3rd request; fetch error
    with pytest.raises(RatesFetchError):
        rates_diff = await rates_cog.compare_posted_rates(
            "http://localhost:5000/bogus-url.txt", output_path
        )

    # 4th request; parse error
    with pytest.raises(RatesProcessError):
        rates_diff = await rates_cog.compare_posted_rates(
            "https://www.google.com", output_path
        )

    # 5th request; write error
    with pytest.raises(RatesWriteError):
        rates_diff = await rates_cog.compare_posted_rates(
            rates_url_1, "/invalid/directory"
        )



import logging
import os

import discord
from discord.ext import commands

import poglink
from poglink.bot import ConfigurableBot
from poglink.config import REQUIRED_VALUES, setup_config
from poglink.utils import setup_argparse

logger = logging.getLogger("poglink")


def cli():
    # Configure logging for CLI usage.
    logging.getLogger().setLevel(level=logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)

    formatter = logging.Formatter("%(asctime)s:%(levelname)s:%(name)s: %(message)s")
    ch.setFormatter(formatter)
    logging.getLogger().addHandler(ch)

    # Define CLI arguments.
    parser = setup_argparse()

    # Load arguments from CLI
    args = parser.parse_args()

    # Override default log level
    if args.debug or os.getenv("BOT_DEBUG", "").lower() == "true":
        ch.setLevel(logging.DEBUG)
        logger.debug("Debug mode activated.")

    # Set up configuration dict
    config_dict = setup_config(args)

    logger.debug(f"Running with config: {config_dict}")

    # Check for required values
    missing_vals = []
    for val in REQUIRED_VALUES:
        if not config_dict.get(val):
            missing_vals.append(val)
    if len(missing_vals) > 0:
        logger.error(f"Missing required configuration variables: {missing_vals}")
        exit(1)

    run(**config_dict)


def run(**kwargs):
    client = ConfigurableBot(".", kwargs)  # passing as dict, not unpacked kwargs

    # load a cog
    @client.command()
    @commands.has_any_role(*client.config.allowed_roles)
    async def load(ctx, extension):
        logger.debug(f"Loading extension: {extension}")
        client.load_extension(f"poglink.cogs.{extension}")

    # unload a cogcd
    @client.command()
    @commands.has_any_role(*client.config.allowed_roles)
    async def unload(ctx, extension):
        logger.debug(f"Unloading extension: {extension}")
        client.unload_extension(f"poglink.cogs.{extension}")

    # reload a cog
    @client.command()
    @commands.has_any_role(*client.config.allowed_roles)
    async def reload(ctx, extension):
        logger.debug(f"Reloading extension: {extension}")
        client.unload_extension(f"poglink.cogs.{extension}")
        client.load_extension(f"poglink.cogs.{extension}")
        await ctx.send(f"Extension '{extension}' reloaded.")

    # load all cogs on boot
    for ext in poglink.cogs.__all__:
        logger.debug(f"Loading extension: {ext}")
        if ext != "bans":  # TODO: Reimplement when bans are enabled
            client.load_extension(f"poglink.cogs.{ext}")

    # run the bot
    try:
        client.run(client.config.token)
    except discord.errors.LoginFailure as e:
        logger.error("Problem with token.")
        exit(1)


if __name__ == "__main__":

    cli()



import datetime

import pytest

from poglink.models import BansStatus


@pytest.fixture()
def sample_bans_txt():
    return """\
Today's Ban Summary
===================
BattlEye Bans: 2160
PC Bans: 9
Xbox Bans: 4
PS4 Bans: 5


7-Day Ban Summary
==================
BattlEye Bans: 13824
PC Bans: 62
Xbox Bans: 84
PS4 Bans: 109


Lifetime Ban Summary
====================
BattlEye Bans: 34218
PC Bans: 15609
Xbox Bans: 13037
PS4 Bans: 15520


Last Updated: 02 Dec 2021 17:59:03 ET
"""


@pytest.fixture()
def sample_bans_dict():
    return {
        "ban_summaries": {
            "Today's Ban Summary": {
                "BattlEye Bans": 2160,
                "PC Bans": 9,
                "Xbox Bans": 4,
                "PS4 Bans": 5,
            },
            "7-Day Ban Summary": {
                "BattlEye Bans": 13824,
                "PC Bans": 62,
                "Xbox Bans": 84,
                "PS4 Bans": 109,
            },
            "Lifetime Ban Summary": {
                "BattlEye Bans": 34218,
                "PC Bans": 15609,
                "Xbox Bans": 13037,
                "PS4 Bans": 15520,
            },
        },
        "last_updated": "2021-12-02T17:59:03",
    }


def test_parse_raw(sample_bans_txt):
    status_dict, last_updated = BansStatus.parse_raw(sample_bans_txt)

    assert status_dict == {
        "Today's Ban Summary": {
            "BattlEye Bans": 2160,
            "PC Bans": 9,
            "Xbox Bans": 4,
            "PS4 Bans": 5,
        },
        "7-Day Ban Summary": {
            "BattlEye Bans": 13824,
            "PC Bans": 62,
            "Xbox Bans": 84,
            "PS4 Bans": 109,
        },
        "Lifetime Ban Summary": {
            "BattlEye Bans": 34218,
            "PC Bans": 15609,
            "Xbox Bans": 13037,
            "PS4 Bans": 15520,
        },
    }
    assert last_updated == datetime.datetime(
        year=2021, month=12, day=2, hour=17, minute=59, second=3
    )


def test_from_raw_to_dict(sample_bans_txt, sample_bans_dict):
    bans_status = BansStatus.from_raw(sample_bans_txt)
    assert bans_status.to_dict() == sample_bans_dict


def test_from_raw_to_raw(sample_bans_txt):

    bans_status = BansStatus.from_raw(sample_bans_txt)
    bans_status_raw = bans_status.to_raw()
    bans_status_from_raw = BansStatus.from_raw(bans_status_raw)

    assert bans_status == bans_status_from_raw


def test_from_dict_to_dict(sample_bans_dict):
    bans_status = BansStatus.from_dict(sample_bans_dict)

    # Check that date objects are converted properly
    assert bans_status.last_updated == datetime.datetime(
        year=2021, month=12, day=2, hour=17, minute=59, second=3
    )

    # Check that the conversion works in reverse
    assert bans_status.to_dict() == sample_bans_dict



import asyncio
import copy
import json
import logging
import os
import re
import time
from json.decoder import JSONDecodeError
from urllib.parse import urlparse

import aiohttp
import discord
from discord.ext import commands

from poglink.config import MIN_POLLING_DELAY
from poglink.error import (
    RatesFetchError,
    RatesProcessError,
    RatesReadError,
    RatesWriteError,
)
from poglink.models import RatesStatus

logger = logging.getLogger(__name__)

EMBED_IMAGE = "https://i.stack.imgur.com/Fzh0w.png"

# create cog class
class Rates(commands.Cog):
    DEFAULT_SERVER_INFO = {
        "smalltribes": {
            "short_name": "Smalltribes",
            "color": 0xA34C44,
        },
        "arkpocalypse": {
            "short_name": "Arkpocalypse",
            "color": 0xF8DE74,
        },
        "conquest": {
            "short_name": "Conquest/Classic",
            "color": 0x6DFF90,
        },
        None: {
            "short_name": "Official",
            "color": 0x63BCC3,
        },
    }

    def __init__(self, client):
        self.client = client

        self.webpage_urls = client.config.rates_urls
        self.channel_id = client.config.rates_channel_id
        self.polling_delay = client.config.polling_delay
        self.allowed_roles = client.config.allowed_roles
        self.data_dir = os.path.expanduser(client.config.data_dir)
        self.output_paths = [
            os.path.join(
                self.data_dir,
                f"last_{os.path.splitext(os.path.basename(urlparse(url).path))[0]}.json",
            )
            for url in client.config.rates_urls
        ]
        logger.debug(f"URLs: {self.webpage_urls}, Output paths: {self.output_paths}")
        # Create parent directory for persistent data if it doesn't exist yet
        if not os.path.exists(self.data_dir):
            logger.info(f"Data directory doesn't exist yet; creating: {self.data_dir}")
            os.makedirs(self.data_dir)

    @staticmethod
    async def get_current_rates(url):
        logger.debug(f"Requesting current server rates")
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url,
                headers={"Pragma": "no-cache", "Cache-Control": "no-cache"},
            ) as response:
                response = await response.text()
                rates = RatesStatus.from_raw(response)
                logger.debug(f"Obtained server rates: {RatesStatus}")
                return rates

    async def send_embed(self, description, url):
        # generate embed
        logger.debug(f"Attempting to send embed. desc: {description}, url: {url}")
        try:
            server_match_dict = (
                re.match(
                    "(?P<host>.*\/)?(?:(?P<platform>.*)\_(?P<game_mode>.*)\_)?dynamicconfig\.ini",
                    os.path.basename(urlparse(url).path),
                )
            ).groupdict()
            server_meta = self.DEFAULT_SERVER_INFO.get(
                server_match_dict.get("game_mode")
            )
            logger.debug(f"Server meta: {server_meta}")
            # TODO: Add ability to accept custom rates URL

        except Exception as e:
            logger.error(f"Rates url could not be processed: {url} {e}")
            return

        server_name = server_meta.get("short_name")

        embed = discord.Embed(
            description=description,
            title=f"ARK's {server_name} server rates have just been updated!",
            color=server_meta.get("color"),
        )
        embed.set_image(url=EMBED_IMAGE)

        # send embed
        channel = self.client.get_channel(self.channel_id)
        message = await channel.send(embed=embed)

        # if in announcement channel, publish message
        if message.channel.type == discord.ChannelType.news:
            logger.info("Announcement channel detected: Publishing message")
            await message.publish()

    async def compare_posted_rates(self, webpage_url, output_path):
        # get current rates from ARK Web API
        logger.info(f"Retrieving current rates at {webpage_url}")
        try:
            rates = await self.get_current_rates(webpage_url)
        except ValueError as e:
            raise RatesProcessError(e) from e
        except Exception as e:
            raise RatesFetchError(e) from e

        # get old rates from file
        try:
            with open(output_path) as f:
                last_rates_dict = json.load(f)
        except (FileNotFoundError, JSONDecodeError) as e:
            logger.warning(f"Problem loading file at {output_path}: {e}")
            last_rates_dict = copy.deepcopy(rates).to_dict()
            try:
                with open(output_path, "w+") as f:
                    json.dump(rates.to_dict(), f, indent=4)
            except Exception as e:
                raise RatesWriteError(e) from e

        last_rates = RatesStatus.from_dict(last_rates_dict)

        # compare rates to last rates
        rates_diff = last_rates.get_diff(rates)

        try:
            if rates_diff.items:
                # save rates to file
                try:
                    with open(output_path, "w+") as f:
                        json.dump(rates.to_dict(), f, indent=4)
                except Exception as e:
                    raise RatesWriteError(e) from e
        except Exception as e:
            raise RatesReadError(e) from e

        return rates_diff

    # Events
    @commands.Cog.listener()
    async def on_ready(self):
        logger.info("Cog Ready: Rates")

        while True:
            for url, output_path in zip(self.webpage_urls, self.output_paths):
                try:
                    rates_diff = await self.compare_posted_rates(url, output_path)
                except RatesFetchError as e:
                    logger.error(
                        f"Could not retrieve rates from ARK Web API at {url}: {e}"
                    )
                except RatesProcessError as e:
                    logger.error(f"Could not process rates URL {url}: {e}")
                except RatesWriteError as e:
                    logger.error(f"Failed to write rates to {output_path}: {e}")
                except Exception as e:
                    logger.error(e)
                try:
                    if rates_diff.items:
                        # generate and send embed
                        logger.info(f"Rates at {url} changed - sending embed")
                        embed_description = rates_diff.to_embed()
                        await self.send_embed(embed_description, url)
                    else:
                        logger.debug(f"No change in rates at {url}.")
                except Exception as e:
                    raise RatesReadError(e) from e

                time.sleep(self.polling_delay)


# add cog to client
def setup(client):
    client.add_cog(Rates(client))



# TODO: Transition to setup.cfg https://setuptools.pypa.io/en/latest/userguide/quickstart.html#transitioning-from-setup-py-to-setup-cfg
from setuptools import setup
import os

# If PYTHON_PACKAGE_VERSION variable is provided, use this instead of automatically grabbing from SCM.
version = os.getenv("PYTHON_PACKAGE_VERSION")

# the local scheme is overridden because pypi doesn't support commit hashes in version numbers
if version:
    use_scm_version = False
elif os.getenv("CI") == "true":
    use_scm_version = {"local_scheme": "no-local-version"}
else:
    use_scm_version = True

setup(
    version=version,
    use_scm_version=use_scm_version,
    setup_requires=["setuptools_scm"],
)



import pytest

from poglink.bot import ConfigurableBot


@pytest.fixture()
def sample_bot(bans_url_1, rates_url_1):
    bot = ConfigurableBot(
        ".",
        {
            "bans_url": bans_url_1,
            "rates_urls": [rates_url_1],
            "data_dir": "tests/data",
        },
    )
    return bot


@pytest.fixture
def configured_httpserver(
    httpserver,
    sample_bansummary_1,
    sample_bansummary_2,
    sample_dynamicconfig_1,
    sample_dynamicconfig_2,
):
    httpserver.expect_request("/bansummary.txt").respond_with_data(
        sample_bansummary_1, content_type="text/plain"
    )
    httpserver.expect_request("/bansummary-changed.txt").respond_with_data(
        sample_bansummary_2, content_type="text/plain"
    )
    httpserver.expect_request("/dynamicconfig.ini").respond_with_data(
        sample_dynamicconfig_1, content_type="text/plain"
    )
    httpserver.expect_request("/dynamicconfig-changed.ini").respond_with_data(
        sample_dynamicconfig_2, content_type="text/plain"
    )

    return httpserver


@pytest.fixture
def rates_url_1(configured_httpserver):
    return configured_httpserver.url_for("/dynamicconfig.ini")


@pytest.fixture
def rates_url_2(configured_httpserver):
    return configured_httpserver.url_for("/dynamicconfig-changed.ini")


@pytest.fixture
def bans_url_1(configured_httpserver):
    return configured_httpserver.url_for("/bansummary.text")


@pytest.fixture
def bans_url_2(configured_httpserver):
    return configured_httpserver.url_for("/bansummary-changed.text")



import argparse


def parse_list(raw):
    """Takes a string representation of a comma separated list and returns a list
    of elements with trimmed whitespace.

    Args:
        raw (str): Comma separated list as a string.

    Returns:
        List[str]: List of trimmed strings.
    """
    return [word.strip() for word in raw.split(",")]


def setup_argparse():
    parser = argparse.ArgumentParser(
        description="""
        Bot for sending messages to Discord whenever ARK Web API changes.
        
        This bot requires some configuration parameters to be set. They can either
        be provided via a config file, CLI args, or environment variables. These
        values are set using the following order of priority:

            1. CLI Arguments
            2. Configuration file
            3. Environment Variables
        """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-d",
        "--data-dir",
        help="Directory in which to store persistent data and config file. Defaults to ~/.poglink",
    )
    parser.add_argument(
        "-t",
        "--token",
        type=str,
        help="API token for ARK. Also can be set via BOT_TOKEN environment variable.",
    )
    parser.add_argument(
        "-p",
        "--polling-delay",
        type=float,
        help="Polling period in seconds. Also can be set via BOT_POLLING_PERIOD environment variable.",
    )
    parser.add_argument(
        "-a",
        "--allowed-roles",
        type=parse_list,
        help="Comma-separated list of allowed roles. Also can be set via BOT_ALLOWED_ROLES environment variable.",
    )
    parser.add_argument(
        "--rates-urls",
        type=parse_list,
        help="Comma-separated list of URLs for ARK rates. Also can be set via BOT_RATES_URLS environment variable.",
    )
    parser.add_argument(
        "--rates-channel-id",
        help="Discord 'rates' channel ID. Also can be set via BOT_RATES_CHANNEL_ID environment variable.",
    )
    # parser.add_argument(
    #     "--bans-url",
    #     help="URL for ARK bans. Also can be set via BOT_BANS_URL environment variable.",
    # ) # TODO: Reimplement when bans are enabled
    # parser.add_argument(
    #     "--bans-channel-id",
    #     help="Discord 'bans' channel ID. Also can be set via BOT_BANS_CHANNEL_ID environment variable.",
    # ) # TODO: Reimplement when bans are enabled
    parser.add_argument("--debug", action="store_true", help="Set log level to DEBUG.")

    return parser
